/************************************************************************/
/*									*/
/*  Ted, Screen drawing and forcing drawing through XClearArea().	*/
/*									*/
/************************************************************************/

#   include	"config.h"

#   include	<stddef.h>
#   include	<stdio.h>

#   include	"tedApp.h"
#   include	"docDraw.h"
#   include	<X11/cursorfont.h>

#   include	<debugon.h>

# define DRDEB(dr) DEB(appDebug( "%s(%3d) %s=[%4d..%4d,%4d..%4d]\n", \
			     __FILE__, __LINE__, #dr, \
			     (dr)->drX0, (dr)->drX1, (dr)->drY0, (dr)->drY1 ))

# define LOG_REDRAWS	0

typedef struct ScreenDrawingData
    {
    int		sddForegroundSet;
#		define	SDDpixelUNDEF	0
#		define	SDDpixelFORE	1
#		define	SDDpixelLINK	2
#		define	SDDpixelBACK	3
#		define	SDDpixelGRID	4
#		define	SDDpixelPAGE	5

    Pixel	sddForePixel;
    Pixel	sddLinkPixel;
    Pixel	sddBackPixel;
    Pixel	sddGridPixel;
    Pixel	sddPagePixel;

    int		sddOx;
    int		sddOy;

    Window	sddWindow;
    GC		sddGc;
    } ScreenDrawingData;

/************************************************************************/
/*									*/
/*  Optimise GC changes.						*/
/*									*/
/************************************************************************/

static void tedDrawSetForegroundFun(	AppDrawingData *	add,
					ScreenDrawingData *	sdd,
					int			which )
    {
    Pixel	pix;

    if  ( sdd->sddForegroundSet == which )
	{ return;	}

    switch( which )
	{
	case SDDpixelFORE:	pix= sdd->sddForePixel;	break;
	case SDDpixelLINK:	pix= sdd->sddLinkPixel;	break;
	case SDDpixelBACK:	pix= sdd->sddBackPixel;	break;
	case SDDpixelGRID:	pix= sdd->sddGridPixel;	break;
	case SDDpixelPAGE:	pix= sdd->sddPagePixel;	break;

	case SDDpixelUNDEF:
	    sdd->sddForegroundSet= which;
	    return;
	default:
	    LDEB(which); return;
	}

    XSetForeground( add->addDisplay, sdd->sddGc, pix );

    sdd->sddForegroundSet= which;
    }

#   define tedDrawSetForeground( add, sdd, which )		\
    if  ( (sdd)->sddForegroundSet != (which) )			\
	{ tedDrawSetForegroundFun( (add), (sdd), (which) );	}

/************************************************************************/
/*									*/
/*  Various Border drawing routines.					*/
/*									*/
/************************************************************************/

static void tedDrawHorizontalBorder(	const BorderProperties *	bp,
					int				asGrid,
					const DrawingContext *		dc,
					ScreenDrawingData *		sdd,
					int				above,
					int				x0,
					int				x1,
					int				y )
    {
    AppDrawingData *	add= dc->dcDrawingData;
    int			wide;
    int			thick= tedBorderThick( &wide, bp, add );

    if  ( asGrid && thick == 0 )
	{ thick= 1;	}

    if  ( thick > 0 )
	{
	const DocumentRectangle *	drClip= dc->dcClipRect;
	DocumentRectangle		drBorder;

	if  ( above )
	    { y -= thick;	}

	drBorder.drX0= x0;
	drBorder.drX1= x1;
	drBorder.drY0= y;
	drBorder.drY1= y+ thick- 1;

	if  ( ! drClip							  ||
	      docIntersectRectangle( &drBorder, &drBorder, dc->dcClipRect ) )
	    {
	    XFillRectangle( add->addDisplay, sdd->sddWindow, sdd->sddGc,
			drBorder.drX0- sdd->sddOx,
			drBorder.drY0- sdd->sddOy,
			drBorder.drX1- drBorder.drX0+ 1,
			drBorder.drY1- drBorder.drY0+ 1 );
	    }
	}

    return;
    }

static void tedDrawVerticalBorder(	const BorderProperties *	bp,
					int				asGrid,
					const DrawingContext *		dc,
					ScreenDrawingData *		sdd,
					int				x,
					int				y0,
					int				y1 )
    {
    AppDrawingData *	add= dc->dcDrawingData;
    int			wide;
    int			thick= tedBorderThick( &wide, bp, add );

    if  ( asGrid && thick == 0 )
	{ thick= 1;	}

    if  ( thick > 0 )
	{
	const DocumentRectangle *	drClip= dc->dcClipRect;
	DocumentRectangle		drBorder;

	drBorder.drX0= x;
	drBorder.drX1= x+ thick- 1;
	drBorder.drY0= y0- 1;
	drBorder.drY1= y1;

	if  ( ! drClip							  ||
	      docIntersectRectangle( &drBorder, &drBorder, dc->dcClipRect ) )
	    {
	    XFillRectangle( add->addDisplay, sdd->sddWindow, sdd->sddGc,
			drBorder.drX0- sdd->sddOx,
			drBorder.drY0- sdd->sddOy,
			drBorder.drX1- drBorder.drX0+ 1,
			drBorder.drY1- drBorder.drY0+ 1 );
	    }
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Draw a text item.							*/
/*									*/
/*  1)  Tabs need not to be drawn.					*/
/*									*/
/************************************************************************/

static int tedDrawParticules(	const BufferItem *	bi,
				const TextParticule *	tp,
				int			count,
				int			baseLine,
				ScreenDrawingData *	sdd,
				DrawingContext *	dc )
    {
    AppDrawingData *	add= dc->dcDrawingData;
    unsigned char *	paraString= bi->biParaString;

    int			drawn;
    int			i;
    int			len;
    int			y;

    AppPhysicalFont *	apf;

    /*  1  */
    switch( tp->tpKind )
	{ 
	case DOCkindTAB:
	    if  ( tp->tpObjectNumber >= 0			&&
		  tp->tpObjectNumber < bi->biParaTabCount	)
		{
		TabStop *	ts= bi->biParaTabStops+ tp->tpObjectNumber;

		switch( ts->tsLeader )
		    {
		    case DOCtlNONE:
			break;

		    case DOCtlDOTS:
			if  ( tp->tpTextAttribute.taFontIsBold )
			    {
			    static char	boldDot[]= { 2, 1 };

			    XSetLineAttributes( add->addDisplay, sdd->sddGc,
				    2, LineOnOffDash, CapButt, JoinMiter );

			    XSetDashes( add->addDisplay, sdd->sddGc,
						0, boldDot, sizeof( boldDot ) );
			    }
			else{
			    static char	dot[]= { 1, 2 };

			    XSetLineAttributes( add->addDisplay, sdd->sddGc,
				    1, LineOnOffDash, CapButt, JoinMiter );

			    XSetDashes( add->addDisplay, sdd->sddGc,
						    0, dot, sizeof( dot ) );
			    }

			XDrawLine( add->addDisplay, sdd->sddWindow, sdd->sddGc,
			    tp->tpX0- sdd->sddOx, baseLine- sdd->sddOy,
			    tp->tpX0+ tp->tpPixelsWide- sdd->sddOx,
			    baseLine- sdd->sddOy );

			break;

		    case DOCtlHYPH:
		    case DOCtlUNDERLINE:
		    case DOCtlTHICK:
		    case DOCtlEQUAL:
		    default:
			LDEB(ts->tsLeader);
		    }
		}

	    drawn= 1;
	    return drawn;

	case DOCkindTEXT:
	    break;

	case DOCkindFIELDSTART:
	case DOCkindFIELDEND:
	case DOCkindBKMKSTART:
	case DOCkindBKMKEND:
	case DOCkindXE:
	case DOCkindTC:
	case DOCkindLINEBREAK:
	case DOCkindPAGEBREAK:
	    drawn= 1;
	    return drawn;

	case DOCkindOBJECT:
	    if  ( tedDrawObject( bi, tp, baseLine, sdd->sddOx, sdd->sddOy,
				add->addDisplay, sdd->sddWindow, sdd->sddGc ) )
		{ LDEB(1); return -1;	}
	    return 1;

	default:
	    LDEB(tp->tpKind); return -1;
	}

    drawn= 1;
    len= tp[drawn-1].tpStroff+ tp[drawn-1].tpStrlen- tp->tpStroff;

    if  ( tp->tpPhysicalFont < 0 )
	{ LDEB(tp->tpPhysicalFont); return -1;	}

    apf= add->addPhysicalFontList.apflFonts+ tp->tpPhysicalFont;
    y= baseLine;

    if  ( tp->tpTextAttribute.taSuperSub == DOCfontSUPERSCRIPT )
	{ y -= ( 3* apf->apfFullSizePixels )/ 10; }

    if  ( tp->tpTextAttribute.taSuperSub == DOCfontSUBSCRIPT )
	{ y += ( 3* apf->apfFullSizePixels )/ 10; }

    if  ( apf->apfFontSet )
	{
	XmbDrawString( add->addDisplay,
		sdd->sddWindow, apf->apfFontSet, sdd->sddGc,
		tp->tpX0- sdd->sddOx, y- sdd->sddOy,
		(char *)paraString+ tp->tpStroff, len );
	}
    else{
	if  ( tp->tpPhysicalFont != dc->dcCurrentPhysicalFont )
	    {
	    XSetFont( add->addDisplay, sdd->sddGc, apf->apfFontStruct->fid );
	    dc->dcCurrentPhysicalFont= tp->tpPhysicalFont;
	    }

	XDrawString( add->addDisplay, sdd->sddWindow, sdd->sddGc,
				    tp->tpX0- sdd->sddOx, y- sdd->sddOy,
				    (char *)paraString+ tp->tpStroff, len );
	}

    i= 0;
    while( i < drawn )
	{
	int	x0;
	int	x1;
	int	y0;
	int	h;

	if  ( ! tp[i].tpTextAttribute.taIsUnderlined )
	    { i++; continue;	}

	x1= x0= tp[i].tpX0;
	y0= baseLine+ apf->apfUnderLinePositionPixels;
	h= apf->apfUnderlineThicknessPixels;

	while( i < drawn && tp[i].tpTextAttribute.taIsUnderlined )
	    { x1= tp[i].tpX0+ tp[i].tpPixelsWide; i++;	}

	XFillRectangle( add->addDisplay, sdd->sddWindow, sdd->sddGc,
				    x0- sdd->sddOx, y0- sdd->sddOy, x1- x0, h );
	}

    return drawn;
    }

static int tedDrawTextLine(	const BufferItem *		bi,
				int				line,
				const FormattingFrame *		ign_ffTwips,
				const FormattingFrame *		ffPixels,
				void *				vsdd,
				const DocumentProperties *	dp,
				DrawingContext *		dc )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;

    const TextLine *		tl= bi->biParaLines+ line;
    const TextParticule *	tp= bi->biParaParticules+ tl->tlFirstParticule;

    AppDrawingData *	add= dc->dcDrawingData;
    int			done;

    done= 0;
    while( done < tl->tlParticuleCount )
	{
	int		drawn;
	int		baseline;

	if  ( tp->tpTextAttribute.taInField )
	    { tedDrawSetForeground( add, sdd, SDDpixelLINK );	}
	else{ tedDrawSetForeground( add, sdd, SDDpixelFORE );	}

	baseline= tl->tlTopPosition.lpYPixels+ tl->tlLineAscentPixels;

	drawn= tedDrawParticules( bi, tp, tl->tlParticuleCount- done,
							baseline, sdd, dc );
	if  ( drawn < 1 )
	    { LDEB(drawn); return -1;	}

	done += drawn; tp += drawn;
	}

    return done;
    }

static int tedLineRectangle(	const BufferItem *		bi,
				AppDrawingData *		add,
				int				line,
				const BufferSelection *		bs,
				const DocumentRectangle *	drClip,
				const FormattingFrame *		ffPixels,
				DocumentRectangle *		drRedraw )
    {
    double		xfac= add->addMagnifiedPixelsPerTwip;
    BufferPosition	bpLineBegin;
    BufferPosition	bpLineEnd;

    const TextLine *	tl= bi->biParaLines+ line;

    DocumentRectangle	drLine;

    const int		mindLine= 1;

    docLinePositions( &bpLineBegin, &bpLineEnd, bi, line );

    drLine.drY0= tl->tlTopPosition.lpYPixels;
    drLine.drY1= tl->tlTopPosition.lpYPixels+
				TWIPStoPIXELS( xfac, tl->tlLineSpacingTwips );

    if  ( docComparePositions( &(bs->bsBegin), &bpLineBegin, mindLine ) < 0 )
	{
	drLine.drX0= ffPixels->ffX0TextLines;

	if  ( bi->biParaFirstIndentTwips < 0 )
	    { drLine.drX0= ffPixels->ffX0FirstLine;	}
	}
    else{ drLine.drX0= bs->bsBegin.bpX;	}

    if  ( docComparePositions( &(bs->bsEnd), &bpLineEnd, mindLine ) > 0 )
	{ drLine.drX1= ffPixels->ffX1TextLines;	}
    else{ drLine.drX1= bs->bsEnd.bpX;		}

    if  ( ! drClip || docIntersectRectangle( drRedraw, drClip, &drLine ) )
	{ return 1;	}
    else{ return 0;	}
    }

static int tedDrawTextReverse(	const BufferItem *		bi,
				int				line,
				const FormattingFrame *		ign_ffTwips,
				const FormattingFrame *		ffPixels,
				void *				vsdd,
				const DocumentProperties *	dp,
				DrawingContext *		dc )
    {
    const BufferSelection *	bs= dc->dcSelection;
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    DocumentRectangle		drRedraw;
    int				done;

    if  ( tedLineRectangle( bi, add, line, bs,
					dc->dcClipRect, ffPixels, &drRedraw ) )
	{
	XRectangle	xr;

	xr.x= drRedraw.drX0- sdd->sddOx;
	xr.y= drRedraw.drY0- sdd->sddOy;
	xr.width= drRedraw.drX1- drRedraw.drX0;
	xr.height= drRedraw.drY1- drRedraw.drY0+ 1;

	XSetClipRectangles( add->addDisplay, sdd->sddGc,
						    0, 0, &xr, 1, Unsorted );

	done= tedDrawTextLine( bi, line, ign_ffTwips, ffPixels, vsdd, dp, dc );
	}
    else{
	TextLine *	tl= bi->biParaLines+ line;

	done= tl->tlParticuleCount;
	}

    return done;
    }

static int tedHighlightBack(	const BufferItem *		bi,
				int				line,
				const FormattingFrame *		ign_ffTwips,
				const FormattingFrame *		ffPixels,
				void *				vsdd,
				const DocumentProperties *	dp,
				DrawingContext *		dc )
    {
    const BufferSelection *	bs= dc->dcSelection;
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    TextLine *			tl= bi->biParaLines+ line;

    DocumentRectangle		drRedraw;

    if  ( tedLineRectangle( bi, add, line, bs,
					dc->dcClipRect, ffPixels, &drRedraw ) )
	{
	XSetForeground( add->addDisplay, sdd->sddGc, sdd->sddBackPixel );

	XFillRectangle( add->addDisplay, sdd->sddWindow, sdd->sddGc,
		drRedraw.drX0- sdd->sddOx,
		drRedraw.drY0- sdd->sddOy,
		drRedraw.drX1- drRedraw.drX0+ 1,
		drRedraw.drY1- drRedraw.drY0+ 1 );
	}

    return tl->tlParticuleCount;
    }

/************************************************************************/
/*									*/
/*  Draw a single paragraph. The Graphical context has been prepared	*/
/*  by the caller.							*/
/*									*/
/*  7)	Do not print below the bottom of fixed height cells. ( Word	*/
/*	does not do so either ).					*/
/*									*/
/************************************************************************/

static int tedDrawParaBottom(	const BufferItem *		bi,
				const BorderProperties *	bp,
				const FormattingFrame *		ign_ffTwips,
				const FormattingFrame *		ffPixels,
				void *				vsdd,
				DrawingContext *		dc )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;
    double			xfac= add->addMagnifiedPixelsPerTwip;
    int				spaceAfterPixels;

    const int			asGrid= 0;

    int				x0;

    spaceAfterPixels= TWIPStoPIXELS( xfac, bi->biParaSpaceAfterTwips );

    x0= ffPixels->ffX0TextLines;
    if  ( bi->biParaFirstIndentTwips < 0 )
	{ x0= ffPixels->ffX0FirstLine;	}

    tedDrawSetForeground( add, sdd, SDDpixelFORE );

    tedDrawHorizontalBorder( bp, asGrid, dc, sdd, 1,
				x0, ffPixels->ffX1TextLines,
				bi->biBelowPosition.lpYPixels-
				spaceAfterPixels -1 );

    return 0;
    }


static int tedDrawParaTop(	const BufferItem *		bi,
				const BorderProperties *	bp,
				const FormattingFrame *		ign_ffTwips,
				const FormattingFrame *		ffPixels,
				void *				vsdd,
				DrawingContext *		dc )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;
    double			xfac= add->addMagnifiedPixelsPerTwip;
    int				spaceBeforePixels;

    const int			asGrid= 0;

    int				x0;

    spaceBeforePixels= TWIPStoPIXELS( xfac, bi->biParaSpaceBeforeTwips );

    x0= ffPixels->ffX0TextLines;
    if  ( bi->biParaFirstIndentTwips < 0 )
	{ x0= ffPixels->ffX0FirstLine;	}

    tedDrawSetForeground( add, sdd, SDDpixelFORE );

    tedDrawHorizontalBorder( bp, asGrid, dc, sdd, 0,
					    x0, ffPixels->ffX1TextLines,
					    bi->biTopPosition.lpYPixels+
					    spaceBeforePixels );
    return 0;
    }

static int tedDrawCellTop(	const BorderProperties *	bp,
				int				asGrid,
				int				ign_x0Twips,
				int				x0Pixels,
				int				ign_x1Twips,
				int				x1Pixels,
				void *				vsdd,
				DrawingContext *		dc,
				const LayoutPosition *		lpTop )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    if  ( asGrid )
	{ tedDrawSetForeground( add, sdd, SDDpixelGRID );	}
    else{ tedDrawSetForeground( add, sdd, SDDpixelFORE );	}

    tedDrawHorizontalBorder( bp, asGrid, dc, sdd, 0,
				    x0Pixels, x1Pixels, lpTop->lpYPixels );

    return 0;
    }

static int tedDrawCellLeft(	const BorderProperties *	bp,
				int				asGrid,
				void *				vsdd,
				DrawingContext *		dc,
				int				x0Twips,
				int				x0Pixels,
				const LayoutPosition *		lpTop,
				const LayoutPosition *		lpBelow )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    if  ( asGrid )
	{ tedDrawSetForeground( add, sdd, SDDpixelGRID );	}
    else{ tedDrawSetForeground( add, sdd, SDDpixelFORE );	}

    tedDrawVerticalBorder( bp, asGrid, dc, sdd, x0Pixels,
				    lpTop->lpYPixels, lpBelow->lpYPixels );

    return 0;
    }

static int tedDrawCellRight(	const BorderProperties *	bp,
				int				asGrid,
				void *				vsdd,
				DrawingContext *		dc,
				int				x1Twips,
				int				x1Pixels,
				const LayoutPosition *		lpTop,
				const LayoutPosition *		lpBelow )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    if  ( asGrid )
	{ tedDrawSetForeground( add, sdd, SDDpixelGRID );	}
    else{ tedDrawSetForeground( add, sdd, SDDpixelFORE );	}

    tedDrawVerticalBorder( bp, asGrid, dc, sdd, x1Pixels,
				    lpTop->lpYPixels, lpBelow->lpYPixels );

    return 0;
    }

static int tedDrawCellBottom(	const BorderProperties *	bp,
				int				asGrid,
				int				ign_x0Twips,
				int				x0Pixels,
				int				ign_x1Twips,
				int				x1Pixels,
				void *				vsdd,
				DrawingContext *		dc,
				const LayoutPosition *		lpBottom )
    {
    ScreenDrawingData *		sdd= (ScreenDrawingData *)vsdd;
    AppDrawingData *		add= dc->dcDrawingData;

    if  ( asGrid )
	{ tedDrawSetForeground( add, sdd, SDDpixelGRID );	}
    else{ tedDrawSetForeground( add, sdd, SDDpixelFORE );	}

    tedDrawHorizontalBorder( bp, asGrid, dc, sdd, 1,
				    x0Pixels, x1Pixels, lpBottom->lpYPixels );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Skip to the next page.						*/
/*									*/
/************************************************************************/

static void tedDrawPageGap(	ScreenDrawingData *		sdd,
				DrawingContext *		dc,
				int				page )
    {
    AppDrawingData *		add= dc->dcDrawingData;

    DocumentRectangle		drBetween;

    drBetween.drX0= add->addBackRect.drX0;
    drBetween.drX1= add->addBackRect.drX1;
    drBetween.drY0= page* add->addPageStepPixels- add->addPageGapPixels;
    drBetween.drY1= page* add->addPageStepPixels;

    if  ( dc->dcClipRect						   &&
	  ! docIntersectRectangle( &drBetween, &drBetween, dc->dcClipRect ) )
	{ return;	}

    tedDrawSetForeground( add, sdd, SDDpixelPAGE );

    XFillRectangle( add->addDisplay, sdd->sddWindow, sdd->sddGc,
					drBetween.drX0- sdd->sddOx,
					drBetween.drY0- sdd->sddOy,
					drBetween.drX1- drBetween.drX0+ 1,
					drBetween.drY1- drBetween.drY0+ 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Draw an I Bar.							*/
/*									*/
/************************************************************************/
#   define	IW	5

int tedDrawIBar(	const BufferPosition *	bp,
			int			ox,
			int			oy,
			Display *		display,
			Window			win,
			GC			gc		)
    {
    int		y0= bp->bpTopPosition.lpYPixels;

    XFillRectangle( display, win, gc,
				bp->bpX- ox, y0- oy, 1, bp->bpY1Pixels- y0 );

    return 0;
    }

int tedExposeIBar(	const EditDocument *	ed,
			const BufferPosition *	bp )
    {
    DocumentRectangle	drIBar;

    drIBar.drX0= bp->bpX- IW;
    drIBar.drX1= bp->bpX+ IW;
    drIBar.drY0= bp->bpTopPosition.lpYPixels;
    drIBar.drY1= bp->bpY1Pixels;

    tedExposeRectangle( ed, &drIBar, /*scrolledX,Y*/ 0,0 );

    return 0;
    }

/************************************************************************/
/*									*/
/*  Draw the eight blocks around a selected object.			*/
/*									*/
/************************************************************************/

#   define	BLOCK	10

static void tedDrawObjectBlocks(	int			x0,
					int			y,
					int			wide,
					int			high,
					Pixel			borderPixel,
					Pixel			bodyPixel,
					const DrawingContext *	dc,
					ScreenDrawingData *	sdd )
    {
    Display *		display= dc->dcDrawingData->addDisplay;
    Window		win= sdd->sddWindow;
    GC			gc= sdd->sddGc;

    XPoint		xp[8];
    int			i;

    /*  bottom  */
    xp[0].x= x0- sdd->sddOx;
    xp[0].y= y- BLOCK- sdd->sddOy;

    xp[1].x= x0 +wide/2- BLOCK/2- sdd->sddOx;
    xp[1].y= y- BLOCK- sdd->sddOy;

    xp[2].x= x0 +wide- BLOCK- sdd->sddOx;
    xp[2].y= y- BLOCK- sdd->sddOy;

    /*  middle */
    xp[3].x= x0- sdd->sddOx;
    xp[3].y= y- high/2- BLOCK/2 -sdd->sddOy;

    xp[4].x= x0 +wide- BLOCK- sdd->sddOx;
    xp[4].y= y- high/2- BLOCK/2 -sdd->sddOy;

    /*  top  */
    xp[5].x= x0- sdd->sddOx;
    xp[5].y= y- high- sdd->sddOy;

    xp[6].x= x0 +wide/2- BLOCK/2- sdd->sddOx;
    xp[6].y= y- high- sdd->sddOy;

    xp[7].x= x0 +wide- BLOCK- sdd->sddOx;
    xp[7].y= y- high- sdd->sddOy;

    XSetForeground( display, gc, bodyPixel );

    for ( i= 0; i < 8; i++ )
	{
	XFillRectangle( display, win, gc, xp[i].x, xp[i].y, BLOCK, BLOCK );
	}

    XSetForeground( display, gc, borderPixel );

    for ( i= 0; i < 8; i++ )
	{
	XDrawRectangle( display, win, gc,
				    xp[i].x, xp[i].y, BLOCK- 1, BLOCK- 1 );
	}

    XDrawRectangle( display, win, gc,
			x0- sdd->sddOx, y- high- sdd->sddOy, wide- 1, high- 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Expose handler for documents.					*/
/*									*/
/*  2)  Clear background.						*/
/*  2b) If the selected area overlaps with the exposed region, draw the	*/
/*	selection background.						*/
/*  3)  Draw text.							*/
/*  4)	Draw page separators.						*/
/*  5)  Draw I bar if necessary.					*/
/*									*/
/************************************************************************/

void tedRedrawRectangle(	Widget			w,
				TedDocument *		td,
				DocumentRectangle *	drClip,
				AppDrawingData *	add,
				AppColors *		ac,
				int			ox,
				int			oy,
				Pixel			selPixel,
				Pixel			linkFore,
				Pixel			tableFrame,
				Window			win,
				GC			gc )
    {
    Display *			display= XtDisplay( w );
    int				screen= DefaultScreen( display );
    Pixel			whitePixel= WhitePixel( display, screen );
    Pixel			forePixel= BlackPixel( display, screen );

    BufferDocument *		bd= td->tdDocument;
    const BufferItem *		docBi= &(bd->bdItem);
    DocumentProperties *	dp= &(bd->bdProperties);

    DrawingContext		dc;
    ScreenDrawingData		sdd;

    const BufferSelection *	bs= &(td->tdSelection);
    const BufferItem *		selItem= (const BufferItem *)0;

    docInitDrawingContext( &dc );

    dc.dcDrawingData= add;
    dc.dcDrawTableGrid= td->tdDrawTableGrid;
    dc.dcClipRect= drClip;

    sdd.sddWindow= win;
    sdd.sddGc= gc;

    sdd.sddOx= ox;
    sdd.sddOy= oy;

    sdd.sddForePixel= forePixel;
    sdd.sddLinkPixel= linkFore;
    sdd.sddGridPixel= tableFrame;
    sdd.sddPagePixel= tableFrame; /* ? */

#   if LOG_REDRAWS
    docLogRectangle( "REDRAW", drClip );
#   endif

    /*  2  */
    XSetForeground( display, gc, whitePixel );

    XFillRectangle( display, win, gc,
		    drClip->drX0- ox, drClip->drY0-oy,
		    drClip->drX1- drClip->drX0+ 1,
		    drClip->drY1- drClip->drY0+ 1 );

    /*  2a  */
    if  ( tedHasSelection( td ) && ! tedHasIBarSelection( td ) )
	{
	selItem= docBi;

	if  ( drClip->drY0 <= bs->bsEnd.bpY1Pixels			&&
	      drClip->drY1 >= bs->bsBegin.bpTopPosition.lpYPixels	)
	    {
	    if  ( bs->bsBegin.bpBi == bs->bsEnd.bpBi	&&
		  ! bs->bsBegin.bpBi->biParaInTable	)
		{ selItem= bs->bsBegin.bpBi;	}

	    if  ( bs->bsBegin.bpBi->biParaInTable		&&
		  docSelectionInsideRow( bs )			)
		{ selItem= bs->bsEnd.bpBi->biParent->biParent; }
	    }
	}

    if  ( selItem )
	{
	tedDrawSetForeground( add, &sdd, SDDpixelUNDEF );
	sdd.sddBackPixel= selPixel;

	dc.dcDrawTextLine= tedHighlightBack;
	dc.dcDrawParaTop= (DRAW_PARA_TOP)0;
	dc.dcDrawParaBottom= (DRAW_PARA_BOTTOM)0;
	dc.dcDrawCellTop= (DRAW_CELL_TOP)0;
	dc.dcDrawCellBottom= (DRAW_CELL_BOTTOM)0;
	dc.dcDrawCellLeft= (DRAW_CELL_LEFT)0;
	dc.dcDrawCellRight= (DRAW_CELL_RIGHT)0;
	dc.dcParaFramePixels= tedParagraphFrame;
	dc.dcFinishPage= (FINISH_PAGE)0;
	dc.dcStartPage= (START_PAGE)0;

	dc.dcLayoutHeader= (LAYOUT_HEADER)0;
	dc.dcLayoutFooter= (LAYOUT_FOOTER)0;
	dc.dcDrawHeadersFooters= 0;

	dc.dcSelection= bs;

	if  ( docDrawItem( selItem, dp, (void *)&sdd, &dc ) )
	    { LDEB(1);	}
	}

    /*  3  */
    sdd.sddForePixel= forePixel;
    sdd.sddBackPixel= whitePixel;

    dc.dcDrawTextLine= tedDrawTextLine;
    dc.dcDrawParaTop= tedDrawParaTop;
    dc.dcDrawParaBottom= tedDrawParaBottom;
    dc.dcDrawCellTop= tedDrawCellTop;
    dc.dcDrawCellBottom= tedDrawCellBottom;
    dc.dcDrawCellLeft= tedDrawCellLeft;
    dc.dcDrawCellRight= tedDrawCellRight;
    dc.dcParaFramePixels= tedParagraphFrame;

    dc.dcLayoutHeader= tedLayoutHeaderItem;
    dc.dcLayoutFooter= tedLayoutFooterItem;
    dc.dcDrawHeadersFooters= 1;

    dc.dcSelection= (BufferSelection *)0;

    XSetBackground( display, gc, whitePixel );

    tedDrawSetForeground( add, &sdd, SDDpixelUNDEF );
    tedDrawSetForeground( add, &sdd, SDDpixelFORE );
    dc.dcCurrentPhysicalFont= -1;

    if  ( docDrawItem( docBi, dp, (void *)&sdd, &dc ) )
	{ LDEB(1);	}

    {
    int		firstPage;
    int		lastPage;

    int		page;
    int		sect;

    int		y0= drClip->drY0;
    int		y1= drClip->drY1;

    y0= drClip->drY0- add->addPageGapPixels;
    if  ( y0 < 0 )
	{ y0= 0;	}

    y1= drClip->drY1+ add->addPageStepPixels- 1;
    if  ( y1 > docBi->biBelowPosition.lpYPixels )
	{ y1=  docBi->biBelowPosition.lpYPixels;	}

    firstPage= y0/ add->addPageStepPixels;
    lastPage=  y1/ add->addPageStepPixels;

    if  ( lastPage > docBi->biBelowPosition.lpPage )
	{ lastPage=  docBi->biBelowPosition.lpPage;	}

    for ( page= firstPage; page <= lastPage; page++ )
	{ tedDrawPageGap( &sdd, &dc, page );	}

    for ( sect= 0; sect < docBi->biGroupChildCount; sect++ )
	{
	BufferItem *	sectBi= docBi->biGroupChildren[sect];
	int		first= sectBi->biTopPosition.lpPage;
	int		last= sectBi->biBelowPosition.lpPage;

	if  ( first < firstPage )
	    { first=  firstPage;	}
	if  ( last > lastPage )
	    { last= lastPage;	}

	for ( page= first; page <= last; page++ )
	    {
	    docDrawPageHeader( sectBi, dp, (void *)&sdd, &dc, page );
	    docDrawPageFooter( sectBi, dp, (void *)&sdd, &dc, page );
	    }
	}

    }

    if  ( selItem && ac->acDepth < 4 )
	{
	sdd.sddForePixel= whitePixel;
	sdd.sddLinkPixel= whitePixel;

	tedDrawSetForeground( add, &sdd, SDDpixelUNDEF );
	tedDrawSetForeground( add, &sdd, SDDpixelFORE );

	dc.dcDrawTextLine= tedDrawTextReverse;
	dc.dcDrawParaTop= (DRAW_PARA_TOP)0;
	dc.dcDrawParaBottom= (DRAW_PARA_BOTTOM)0;
	dc.dcDrawCellTop= (DRAW_CELL_TOP)0;
	dc.dcDrawCellBottom= (DRAW_CELL_BOTTOM)0;
	dc.dcDrawCellLeft= (DRAW_CELL_LEFT)0;
	dc.dcDrawCellRight= (DRAW_CELL_RIGHT)0;
	dc.dcParaFramePixels= tedParagraphFrame;
	dc.dcFinishPage= (FINISH_PAGE)0;
	dc.dcStartPage= (START_PAGE)0;

	dc.dcLayoutHeader= (LAYOUT_HEADER)0;
	dc.dcLayoutFooter= (LAYOUT_FOOTER)0;
	dc.dcDrawHeadersFooters= 0;

	dc.dcSelection= bs;

	if  ( docDrawItem( selItem, dp, (void *)&sdd, &dc ) )
	    { LDEB(1);	}
	}

    XSetForeground( display, gc, sdd.sddForePixel );

    {
    TextLine *		tl;
    double		xfac= add->addMagnifiedPixelsPerTwip;

    /*  5  */
    if  ( tedHasIBarSelection( td ) )
	{
	const BufferPosition *	bpBegin= &(td->tdSelection.bsBegin);

	if  ( drClip->drY0 <= bpBegin->bpY1Pixels			&&
	      drClip->drY1 >= bpBegin->bpTopPosition.lpYPixels		&&
	      ! td->tdShowIBarId			)
	    { tedDrawIBar( bpBegin, ox, oy, display, win, gc ); }
	}
    else{
	BufferItem *		bi;
	TextParticule *		tp;
	InsertedObject *	io;

	if  ( td->tdObjectSelected					&&
	      ! tedGetObjectSelection( td, &bi, &tl, &tp, &io  )	)
	    {
	    int			y1;

	    y1= tl->tlTopPosition.lpYPixels+
				TWIPStoPIXELS( xfac, tl->tlLineSpacingTwips );

	    if  ( drClip->drY0 <= y1					&&
		  drClip->drY1 >= tl->tlTopPosition.lpYPixels		)
		{
		int			baseline;

		baseline= tl->tlTopPosition.lpYPixels+ tl->tlLineAscentPixels;

		if  ( io->ioDragWide > 0 )
		    {
		    tedDrawObjectBlocks( tp->tpX0,
			    baseline+ io->ioDragHigh- io->ioPixelsHigh,
			    io->ioDragWide, io->ioDragHigh,
			    sdd.sddForePixel, selPixel, &dc, &sdd );
		    }
		else{
		    tedDrawObjectBlocks( tp->tpX0, baseline,
			    io->ioPixelsWide, io->ioPixelsHigh,
			    sdd.sddForePixel, selPixel, &dc, &sdd );
		    }
		}
	    }
	}
    }
	  
    return;
    }

/************************************************************************/
/*									*/
/*  Cause a rectangle to be redrawn.					*/
/*									*/
/*  Cause the smallest rectangle that contains the selection to be	*/
/*  redrawn.								*/
/*									*/
/************************************************************************/

void tedExposeRectangle(	const EditDocument *		ed,
				const DocumentRectangle *	drChanged,
				int				scrolledX,
				int				scrolledY )
    {
    Widget			w= ed->edDocumentWidget;
    Display *			display= XtDisplay( w );
    Window			win= XtWindow( w );

    DocumentRectangle		drExpose;
    DocumentRectangle		drScrolled;

    if  ( ! drChanged )
	{ XClearArea( display, win, 0, 0, 0, 0, True ); return; }

    drScrolled= *drChanged;
    drScrolled.drX0 -= scrolledX;
    drScrolled.drX1 -= scrolledX;
    drScrolled.drY0 -= scrolledY;
    drScrolled.drY1 -= scrolledY;

    if  ( scrolledX != 0 )
	{
	drScrolled.drY0= ed->edVisibleRect.drY0;
	drScrolled.drY1= ed->edVisibleRect.drY1;
	}

    if  ( scrolledY != 0 )
	{
	drScrolled.drX0= ed->edVisibleRect.drX0;
	drScrolled.drX1= ed->edVisibleRect.drX1;
	}

    docUnionRectangle( &drScrolled, &drScrolled, drChanged );

    drExpose.drX0= ed->edVisibleRect.drX0;
    drExpose.drY0= ed->edVisibleRect.drY0;
    drExpose.drX1= ed->edVisibleRect.drX1;
    drExpose.drY1= ed->edVisibleRect.drY1;

    if  ( docIntersectRectangle( &drExpose, &drExpose, &drScrolled ) )
	{
	int	ox= ed->edVisibleRect.drX0;
	int	oy= ed->edVisibleRect.drY0;

#	if LOG_REDRAWS
	docLogRectangle( "CLEAR!", &drExpose );
#	endif

	XClearArea( display, win,
		drExpose.drX0- ox,
		drExpose.drY0- oy,
		drExpose.drX1- drExpose.drX0+ 1,
		drExpose.drY1- drExpose.drY0+ 1, True );
	}
    }

void tedExposeSelection(	const EditDocument *	ed,
				const BufferSelection *	bs,
				int			scrolledX,
				int			scrolledY )
    {
    const AppDrawingData *	add= &(ed->edDrawingData);
    DocumentRectangle		drClip;

    tedSelectionRectangle( &drClip, add, bs );

    tedExposeRectangle( ed, &drClip, scrolledX, scrolledY );

    return;
    }

/************************************************************************/
/*									*/
/*  Blinking cursor.							*/
/*									*/
/*  1)  Turned off when we are debugging exposures and redraws.		*/
/*									*/
/************************************************************************/

void tedUndrawIBar(	const EditDocument *	ed )
    {
#   if LOG_REDRAWS
    /*  1  */
#   else
    TedDocument *	td= (TedDocument *)ed->edPrivateData;

    tedExposeIBar( ed, &(td->tdSelection.bsBegin) );
#   endif

    return;
    }

void tedRedrawIBar(	TedDocument *	td,
			int		ox,
			int		oy,
			Display *	display,
			Window		win,
			GC		gc )
    {
    tedDrawIBar( &(td->tdSelection.bsBegin), ox, oy, display, win, gc );

    return;
    }

/************************************************************************/
/*									*/
/*  Create the windows that are used to reshape selected objects.	*/
/*									*/
/************************************************************************/

void tedSetObjectWindows(	TedDocument *		td,
				BufferItem *		bi,
				TextLine *		tl,
				TextParticule *		tp,
				int			ox,
				int			oy,
				Display *		display,
				Window			win )
    {
    InsertedObject *		io= bi->biParaObjects+ tp->tpObjectNumber;
    int				wide= io->ioPixelsWide;
    int				high= io->ioPixelsHigh;

    XSetWindowAttributes	xswa;

    static Cursor		moveCursor;
    static Cursor		bottomCursor;
    static Cursor		rightCursor;
    static Cursor		cornerCursor;

    int				x0= tp->tpX0;
    int				baseline;

    baseline= tl->tlTopPosition.lpYPixels+ tl->tlLineAscentPixels;

    if  ( ! moveCursor )
	{
	moveCursor= XCreateFontCursor( display, XC_fleur );
	bottomCursor= XCreateFontCursor( display, XC_bottom_side );
	rightCursor= XCreateFontCursor( display, XC_right_side );
	cornerCursor= XCreateFontCursor( display, XC_bottom_right_corner );
	}

    if  ( ! td->tdObjectWindow )
	{
	xswa.cursor= moveCursor;
	td->tdObjectWindow= XCreateWindow( display, win,
			    x0- ox, baseline- high- oy,
			    wide, high,
			    0, CopyFromParent, InputOnly, CopyFromParent,
			    0L, &xswa );
			    /*
			    CWCursor, &xswa );
			    */

	xswa.cursor= bottomCursor;
	td->tdObjectBottomWindow= XCreateWindow( display, td->tdObjectWindow,
			    wide/ 2- BLOCK/ 2, high- BLOCK,
			    BLOCK, BLOCK,
			    0, CopyFromParent, CopyFromParent, CopyFromParent,
			    CWCursor, &xswa );

	xswa.cursor= rightCursor;
	td->tdObjectRightWindow= XCreateWindow( display, td->tdObjectWindow,
			    wide- BLOCK, high/ 2- BLOCK/ 2,
			    BLOCK, BLOCK,
			    0, CopyFromParent, CopyFromParent, CopyFromParent,
			    CWCursor, &xswa );

	xswa.cursor= cornerCursor;
	td->tdObjectCornerWindow= XCreateWindow( display, td->tdObjectWindow,
			    wide- BLOCK, high- BLOCK,
			    BLOCK, BLOCK,
			    0, CopyFromParent, CopyFromParent, CopyFromParent,
			    CWCursor, &xswa );

	XMapRaised( display, td->tdObjectBottomWindow );
	XMapRaised( display, td->tdObjectRightWindow );
	XMapRaised( display, td->tdObjectCornerWindow );
	}
    else{
	XMoveResizeWindow( display, td->tdObjectWindow,
				x0- ox, baseline- high- oy, wide, high );

	XMoveWindow( display, td->tdObjectBottomWindow,
				wide/ 2- BLOCK/ 2, high- BLOCK );
	XMoveWindow( display, td->tdObjectRightWindow,
				wide- BLOCK, high/ 2- BLOCK/ 2 );
	XMoveWindow( display, td->tdObjectCornerWindow,
				wide- BLOCK, high- BLOCK );
	}

    XMapRaised( display, td->tdObjectWindow );

    td->tdObjectSelected= 1;
    }
